#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsApplicable.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 27.02.2020
# Last Modified Date: 09.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any
from ampel.base import abstractmethod, AmpelBaseModel
from ampel.base.AmpelABC import AmpelABC

class AbsApplicable(AmpelABC, AmpelBaseModel, abstract=True):

	@abstractmethod
	def apply(self, arg: Any) -> Any:
		...
